import { Socket } from 'socket.io';
import { DefaultEventsMap } from 'socket.io/dist/typed-events';

import { createClient, LiveTranscriptionEvents } from '@deepgram/sdk';

import dotenv from 'dotenv';
import { RealtimeUtils } from '@openai/realtime-api-beta';
import { ElevenLabsClient } from 'elevenlabs';
import { OptimizeStreamingLatency } from 'elevenlabs/api';
import { Writable } from 'stream';
import Cartesia from '@cartesia/cartesia-js';
import WS from 'ws';
import OpenAI from 'openai';
import { ChatCompletionMessageParam } from 'openai/resources';

import elevenlabs_wss from './elevenlabs_wss';
import fs from 'fs';

dotenv.config();

import { AssemblyAI, RealtimeTranscript } from 'assemblyai';

const assemblyAI = new AssemblyAI({
    apiKey: process.env.ASSEMBLYAI_API_KEY || '',
});

const cartesia = new Cartesia({
    apiKey: process.env.CARTESIA_API_KEY,
});

const openai = new OpenAI({
    apiKey: process.env.OpenAI_API_KEY,
});

// @ts-ignore
global.WebSocket = WS;

export function voiceEngineSocketUser(socket: Socket<DefaultEventsMap, DefaultEventsMap, DefaultEventsMap, any>) {
    let transcript = '';
    let currentSpeachTime = 0;
    let currentSpeach: {
        charStartTimesMs: number[];
        chars: string[];
    }[] = [];

    function logTranscript(prefix: string, text: string) {
        transcript += prefix + ':\t' + text + '\n';
    }

    console.log('A user connected');

    socket.on('bot-start', async () => {
        let userMessages: ChatCompletionMessageParam[] = [
            {
                role: 'system',
                content: `Telefonvertriebler bei Sentrovo
Rolle Du bist Kevin von Sentrovo, einer KI-Agentur, die spezialisierte KI-Dienstleistungen anbietet. Deine Hauptaufgabe besteht darin, potenzielle Kunden durch Outbound-Anrufe zu erreichen und ein Kennenlerngespräch für ein KI-Audit zu vereinbaren. Du kennst die typischen Vorteile von KI-Lösungen für Unternehmen und kannst anschaulich erklären, wie ein Audit dabei hilft, Potenziale zu identifizieren.

Heutiges Datum: {{now}}
Informationen Im Folgenden sind die Informationen über die Person, die du anrufst:

Vorname: {{Vorname}}
Nachname: {{Nachname}}
Unternehmen: {{Unternehmen}}
Position: {{Position}}
Verhalten

Sprache: Du sprichst ausschließlich Deutsch.
Freundlichkeit und Umgangston: Sei stets freundlich, zuvorkommend und locker. Verwende gelegentlich umgangssprachliche Phrasen wie „Ähm...“, „Also...“ und „Ich meine...“.
Höflichkeitsform: Sieze den Nutzer.
Kommunikationsform: Deine Antworten sind kurz und prägnant, wie bei einem realen Telefongespräch.
Witz und Charme: Sei charmant und gerne humorvoll, um das Gespräch angenehm zu gestalten.
Aussprache von Daten: Verwende ausgeschriebene Datumsformen, z. B. „am neunundzwanzigsten September“, um eine natürliche Sprache zu gewährleisten.
Anweisungen

Ziel ist es, dass du so wenig wie möglich redest und den Anrufer reden lässt; fokussiere dich darauf.
Fokus auf Mehrwert: Stelle den Nutzen des KI-Audits klar heraus und vermeide unnötige technische Details. Betone die konkreten Vorteile wie Effizienzsteigerungen und Kostensenkungen.
Einsatz von Social Proof: Erwähne gegebenenfalls ähnliche Unternehmen, die bereits positive Erfahrungen gemacht haben, um Vertrauen zu stärken.
Timing für Terminvereinbarung: Bitte immer Termine innerhalb der nächsten zwei Tage an, mit zwei Uhrzeitoptionen. Vermeide dabei das Wochenende.
Positiv abschließen: Nachdem der Termin vereinbart ist, erinnere den Kunden daran, dass Kevin zur vereinbarten Uhrzeit wieder anruft.
Umgang mit Einwänden:

Gefühle des Ansprechpartners bestätigen: Beginne damit, das Gefühl des Ansprechpartners zu bestätigen. Zeige Verständnis, ohne sofort zu widersprechen.
Klarheit schaffen: Stelle eine klärende Frage, die den Nutzen abzieht, den der Kunde erreichen möchte, oder auf das Problem, das er lösen will.
Beispiel: „Ich verstehe, dass Sie Bedenken haben. Könnte ich fragen, welcher Aspekt des KI-Audits Ihnen aktuell am wichtigsten erscheint?“
Reale Einwände isolieren: Verwende gezielte Fragen des SPIN-Frameworks, um zum Kern des Einwands zu gelangen.
Beispiel: „Sie sagen, dass XYZ wichtig ist, weil es Ihnen hilft, Zeit zu sparen – was würde Sie dann davon abhalten, das Audit zu investieren?“
Reale Einwände ansprechen oder minimieren: Je nach Art des Einwands sprich ihn direkt an oder indirekt, warum er weniger relevant ist. Beispielsweise kannst du betonen, wie gering der Aufwand oder das Risiko ist.
Halte deine Antwort kurz und prägnant, du kannst auch warten, bis er es von sich selbst betrachtet.
Skript

Einleitung

„Ich rufe an, um kurz über KI-Möglichkeiten für Ihr Unternehmen zu sprechen. Hätten Sie kurz Zeit?“
Strategie: Erst damit beginnen, sobald du mit dem Entscheider sprichst.
Interesse wecken

„Gibt es Bereiche bei Ihnen, wo Sie sich mehr Effizienz oder weniger Aufwand wünschen würden? KI kann da oft schon sehr hilfreich sein.“
Ziel: Den Ansprechpartner dazu bringen, über mögliche Herausforderungen nachzudenken, die KI lösen könnte.
Nutzen klar und unverbindlich darstellen

„Wir bieten ein KI-Audit an, das Unternehmen hilft, risikofreie Potenziale zu entdecken und besser zu nutzen – ganz unkompliziert und ohne Verpflichtung.“
Taktik: Zeige den Mehrwert des Audits als einfachen, risikofreien Einstieg.
Einwandbehandlung: Kein Interesse / Kein Bedarf

„Total verständlich. Viele unserer Kunden waren anfangs auch skeptisch und haben trotzdem Potenziale gefunden. Ein kurzes Gespräch schadet nicht und könnte schon interessante Ansätze zeigen.“
Strategie: Unverbindlichkeit betonen und das Gespräch als lockere Option präsentieren.
Einwandbehandlung: Kein Budget

„Kein Problem, das Audit ist erstmal kostenfrei. Es geht darum, zusammen zu schauen, ob Potenziale für Einsparungen oder Effizienzgewinne da sind – völlig unverbindlich.“
Technik: Die kostenfreie Option als einfache Möglichkeit ohne Risiko darstellen.
Einwandbehandlung: Entscheidungs­träger nicht verfügbar

„Verstehe ich. Oft ist es hilfreich, ein paar erste Infos zu haben, um das Thema intern anzuschieben. Ich kann Ihnen gern kurz zeigen, wie unser Ansatz funktioniert.“
Strategie: Den Gesprächspartner unterstützen, Informationen für interne Entscheidungsprozesse zu sammeln.
Terminvereinbarung

„Lassen Sie uns doch einen kurzen Termin in den nächsten Tagen ansetzen. Wie würde Ihnen [Datum, z.B. morgen oder übermorgen] um 10:00 Uhr oder 14:00 Uhr passen? Wir können in 15 Minuten alles Wichtige durchgehen.“
Taktik: Zwei konkrete Zeitvorschläge in den nächsten Tagen machen, um die Terminfindung zu erleichtern und sicherzustellen, dass es zeitnah stattfindet.
Bestätigung und Abschluss

„Super, ich habe den Termin am [Datum] um [Uhrzeit] für uns eingetragen. Ich freue mich auf das Gespräch und bedanke mich für Ihre Zeit. Bis dann!“
Taktik: Positiv und locker abschließen, um den Termin als einfach und unverbindlich zu präsentieren.
Hinweis: Der Ton bleibt freundlich und entspannt, um Vertrauen aufzubauen und die Hemmschwelle für ein Kennenlerngespräch zu senken.`,
            },
        ];

        const vars: { [key: string]: string } = {
            Vorname: 'Max',
            Nachname: 'Mustermann',
            Unternehmen: 'Musterfirma',
            Position: 'Geschäftsführer',
            now: new Date().toLocaleDateString(),
        };

        for (const message of userMessages) {
            if (message.content) {
                for (const key in vars) {
                    if (message.content && message.role === 'system') {
                        if (typeof message.content === 'string') {
                            message.content = message.content.replace(new RegExp(`{{${key}}}`, 'g'), vars[key as keyof typeof vars]);
                        }
                    }
                }
            }
        }

        console.log('Bot started');

        async function speakText(text: string) {
            console.log('Generated message:', text);

            const time = new Date().getTime();
            let lastTime = 0;
            let firstMessage = true;
            const labs11 = await elevenlabs_wss.connect(
                {
                    voice_id: 'N2lVS1w4EtoT3dr4eOWO',
                    model_id: 'eleven_turbo_v2_5',

                    optimize_streaming_latency: 4,
                    output_format: 'pcm_24000',
                    language_code: 'de',
                    //sync_alignment: true,
                    inactivity_timeout: 20,
                },
                (data, randomUUID) => {
                    if (!data.audio) return;

                    const audioBuffer = Buffer.from(data.audio, 'base64');

                    const audioBufferArray = new Uint8Array(audioBuffer);

                    socket.emit('openai-audio', audioBufferArray, randomUUID);
                    console.log('Received audio data from Eleven Labs');

                    if (data.normalizedAlignment) {
                        if (firstMessage) {
                            firstMessage = false;
                            currentSpeachTime = new Date().getTime();
                            currentSpeach = [];
                        }

                        currentSpeach.push(data.normalizedAlignment);
                    }
                }
            );

            const ws11 = labs11.socket;

            socket.emit('openai-audio-start', labs11.randomUUID);

            console.log('Connected to Eleven Labs. Took', new Date().getTime() - time, 'ms');

            elevenlabs_wss.generate(ws11, {
                flush: true,
                text: text + ' ',

                voice_settings: {
                    style: 0.5,
                    use_speaker_boost: true,
                    stability: 0.5,
                    similarity_boost: 0.5,
                },
            });
            elevenlabs_wss.generate(ws11, {
                //flush: true,
                text: '',
            });
        }

        async function generateVoiceMessage() {
            const output = await openai.beta.chat.completions.parse({
                model: 'gpt-4o',
                temperature: 0.8,
                //max_completion_tokens: 100,
                messages: [...userMessages],
            });

            const text = output.choices[0].message.content;

            if (!text) return;

            speakText(text);
        }

        function addMessageToUser(message: ChatCompletionMessageParam) {
            userMessages.push(message);
        }

        function addLastMessageToChat() {
            let oldText = '';
            let lastMs = 0;

            const speakOffset = new Date().getTime() - currentSpeachTime;

            let inrerrupt = false;

            for (const alignment of currentSpeach) {
                let index = 0;
                for (const char of alignment.chars) {
                    const ms = alignment.charStartTimesMs[index];

                    if (lastMs + ms < speakOffset) {
                        oldText += char;
                    } else {
                        inrerrupt = true;
                        break;
                    }

                    index++;
                }

                lastMs = alignment.charStartTimesMs[alignment.charStartTimesMs.length - 1];
            }

            if (inrerrupt) {
                oldText += ' ... **ABGEBROCHEN**';
            }

            if (oldText) {
                addMessageToUser({
                    role: 'assistant',
                    content: oldText,
                });
            }

            logTranscript('AI', oldText);
        }

        speakText('Hallo hier ist Kevin von Sentrovo, spreche ich hier mit Herr Mustermann?');

        const transcriber = assemblyAI.realtime.transcriber({
            sampleRate: 16_000,
            encoding: 'pcm_s16le',
        });

        transcriber.on('open', ({ sessionId }) => {
            console.log(`Session opened with ID: ${sessionId}`);
        });

        transcriber.on('error', (error: Error) => {
            console.error('Error:', error);
        });

        transcriber.on('close', (code: number, reason: string) => console.log('Session closed:', code, reason));

        transcriber.on('transcript', (transcript: RealtimeTranscript) => {
            if (!transcript.text) {
                return;
            }

            if (transcript.message_type === 'PartialTranscript') {
                console.log('Partial:', transcript.text);
            } else {
                console.log('Final:', transcript.text);
            }
        });

        console.log('Connecting to real-time transcript service');
        await transcriber.connect();

        console.log('Starting recording');

        /*
         {
                //data.speech_final
                if (data.is_final) {
                    const text = data.channel.alternatives[0].transcript;
                    const confedence = data.channel.alternatives[0].confidence;
                    console.log('Final\t', text);

                    // check if the text is empty
                    if (!text) return;

                    // Send the text to the user
                    socket.emit('bot-text', text);

                    socket.emit('bot-stopSpeaking');

                    addLastMessageToChat();

                    addMessageToUser({
                        role: 'user',
                        content: text,
                    });

                    generateVoiceMessage();

                    logTranscript('User', text);

                    console.log('Transcript:\n', transcript);

                    // Send the text to the GPT
                }
                //console.log('ts\t', data);
            } */

        // STEP 4: Fetch the audio stream and send it to the live transcription connection

        socket.on('bot-voice-data', (audioBuffer: any) => {
            // Convert Buffer to ArrayBuffer
            const arrayBuffer = audioBuffer.buffer.slice(audioBuffer.byteOffset, audioBuffer.byteOffset + audioBuffer.byteLength);

            // Convert ArrayBuffer to Int16Array
            const float32Array = new Float32Array(arrayBuffer);

            //console.log('Received audio data from User:', data);
            //dgConnection.send(RealtimeUtils.floatTo16BitPCM(float32Array));
            //transcriber.stream()

            const audioBufferArray = RealtimeUtils.floatTo16BitPCM(float32Array);

            transcriber.sendAudio(audioBufferArray);
        });

        function stopConversation() {
            console.log('Ending conversation');
            addLastMessageToChat();

            //dgConnection.disconnect();
            socket.emit('bot-stopSpeaking');

            // write the transcript to the file
            const uuid = new Date().getTime();

            // create folder
            fs.mkdirSync(`transcripts/${uuid}`, { recursive: true });

            fs.writeFileSync(`transcripts/${uuid}/transcript.txt`, transcript);
        }

        socket.on('bot-end', () => {
            stopConversation();
        });

        socket.on('disconnect', () => {
            stopConversation();
            console.log('A user disconnected');
        });
    });
}
